"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GenericMap = exports.visitors = exports.walkClass = void 0;
const extends_1 = require("./extends");
const helpers_1 = require("./helpers");
const parser_1 = require("./parser");
const types_1 = require("./types");
// --------------------------------------------------------------------- //
// -------------------------- VISITOR HELPERS -------------------------- //
// --------------------------------------------------------------------- //
function getDecorators(targetClass, targetType, target, index) {
    const natives = Reflect.getOwnMetadata(types_1.DECORATOR_KEY, targetClass) || [];
    const result = [];
    for (let _a of natives) {
        const { allowMultiple, inherit } = _a, item = __rest(_a, ["allowMultiple", "inherit"]);
        const par = item;
        if (item.targetType === targetType && item.target === target && (index == undefined || par.targetIndex === index)) {
            result.push(Object.assign(Object.assign({}, item.value), { [types_1.DecoratorOptionId]: { allowMultiple, inherit } }));
        }
    }
    return result;
}
function getTypeOverrideFromDecorator(decorators) {
    const getType = (type) => typeof type === "object" ? helpers_1.createClass({ definition: type }) : type;
    const override = decorators.find((x) => x.kind === "Override");
    if (!override)
        return;
    // extract type from the callback
    const rawType = helpers_1.metadata.isCallback(override.type) ? override.type({}) : override.type;
    return { type: Array.isArray(rawType) ? [getType(rawType[0])] : getType(rawType), genericParams: override.genericParams };
}
class GenericMap {
    constructor(types) {
        this.maps = [];
        this.maps = this.createMaps(types);
    }
    createMaps(types) {
        const result = [];
        for (const type of types) {
            const parent = Object.getPrototypeOf(type);
            const templates = this.getTemplates(parent);
            if (!templates)
                throw new Error(`Configuration Error: ${parent.name} uses string template type @reflect.type(<string>) but doesn't specify @generic.template()`);
            const types = this.getTypes(type);
            if (!types)
                throw new Error(`Configuration Error: ${type.name} inherit from generic class but doesn't use @generic.type()`);
            if (templates.length !== types.length)
                throw new Error(`Configuration Error: Number of parameters mismatch between @generic.template() on ${parent.name} and @generic.type() on ${type.name}`);
            result.unshift(new Map(templates.map((x, i) => ([x, types[i]]))));
        }
        return result;
    }
    getTemplates(target) {
        const decorator = getDecorators(target, "Class", target.name)
            .find((x) => x.kind === "GenericTemplate");
        return decorator === null || decorator === void 0 ? void 0 : decorator.templates;
    }
    getTypes(target) {
        const decorator = getDecorators(target, "Class", target.name)
            .find((x) => x.kind === "GenericType");
        return decorator === null || decorator === void 0 ? void 0 : decorator.types;
    }
    get(rawType) {
        const isArray = Array.isArray(rawType);
        const type = isArray ? rawType[0] : rawType;
        const result = this.maps.reduce((val, map) => {
            // keep looking at the real type
            // if it is string then it still a generic type template
            return typeof val === "string" ? map.get(val) : val;
        }, type);
        return isArray ? [result] : result;
    }
}
exports.GenericMap = GenericMap;
// --------------------------------------------------------------------- //
// ------------------------- PURIFIER VISITORS ------------------------- //
// --------------------------------------------------------------------- //
var visitors;
(function (visitors) {
    function addsDesignTypes(meta, ctx) {
        const getType = (type, i) => type[i] === Array ? [Object] : type[i];
        if (meta.kind === "Method") {
            const returnType = Reflect.getOwnMetadata(types_1.DESIGN_RETURN_TYPE, ctx.target.prototype, meta.name);
            return Object.assign(Object.assign({}, meta), { returnType });
        }
        else if (helpers_1.metadata.isParameterProperties(meta)) {
            const parTypes = Reflect.getOwnMetadata(types_1.DESIGN_PARAMETER_TYPE, ctx.target) || [];
            return Object.assign(Object.assign({}, meta), { type: getType(parTypes, meta.index) });
        }
        else if (meta.kind === "Property") {
            const type = Reflect.getOwnMetadata(types_1.DESIGN_TYPE, ctx.target.prototype, meta.name);
            return Object.assign(Object.assign({}, meta), { type });
        }
        else if (meta.kind === "Parameter" && ctx.parent.kind === "Constructor") {
            const parTypes = Reflect.getOwnMetadata(types_1.DESIGN_PARAMETER_TYPE, ctx.target) || [];
            return Object.assign(Object.assign({}, meta), { type: getType(parTypes, meta.index) });
        }
        else if (meta.kind === "Parameter" && ctx.parent.kind === "Method") {
            const parTypes = Reflect.getOwnMetadata(types_1.DESIGN_PARAMETER_TYPE, ctx.target.prototype, ctx.parent.name) || [];
            return Object.assign(Object.assign({}, meta), { type: getType(parTypes, meta.index) });
        }
        else
            return meta;
    }
    visitors.addsDesignTypes = addsDesignTypes;
    function addsDecorators(meta, ctx) {
        if (meta.kind === "Parameter" || helpers_1.metadata.isParameterProperties(meta)) {
            const targetName = meta.kind === "Parameter" ? ctx.parent.name : "constructor";
            const decorators = getDecorators(ctx.target, "Parameter", targetName, meta.index);
            return Object.assign(Object.assign({}, meta), { decorators: meta.decorators.concat(decorators) });
        }
        else if (meta.kind === "Method" || meta.kind === "Property" || meta.kind === "Class") {
            const decorators = getDecorators(ctx.target, meta.kind, meta.name);
            return Object.assign(Object.assign({}, meta), { decorators: meta.decorators.concat(decorators) });
        }
        return meta;
    }
    visitors.addsDecorators = addsDecorators;
    function addsTypeOverridden(meta, ctx) {
        if (meta.kind === "Constructor" || meta.kind === "Class")
            return meta;
        const overridden = getTypeOverrideFromDecorator(meta.decorators);
        if (!overridden)
            return meta;
        if (meta.kind === "Method")
            return Object.assign(Object.assign({}, meta), { returnType: overridden.type });
        return Object.assign(Object.assign({}, meta), { type: overridden.type });
    }
    visitors.addsTypeOverridden = addsTypeOverridden;
    function addsGenericOverridden(meta, ctx) {
        const isGeneric = (decorator) => {
            const singleType = Array.isArray(decorator.type) ? decorator.type[0] : decorator.type;
            return typeof singleType === "string" || decorator.genericParams.length > 0;
        };
        const isString = (decorator) => {
            const singleType = Array.isArray(decorator.type) ? decorator.type[0] : decorator.type;
            return typeof singleType === "string";
        };
        const getGenericType = (map, decorator) => {
            const converted = [];
            for (const param of decorator.genericParams) {
                converted.push(map.get(param));
            }
            return helpers_1.createClass({ parent: decorator.type, genericParams: converted });
        };
        if (meta.kind === "Constructor" || meta.kind === "Class")
            return meta;
        const decorator = getTypeOverrideFromDecorator(meta.decorators);
        if (!decorator || !decorator.type || !isGeneric(decorator))
            return meta;
        const map = new GenericMap(ctx.classPath);
        const type = isString(decorator) ? map.get(decorator.type) : getGenericType(map, decorator);
        // if current class has @generic.template() then process
        if (meta.kind === "Method") {
            // if type is not a generic template type then return immediately
            return Object.assign(Object.assign({}, meta), { returnType: type });
        }
        return Object.assign(Object.assign({}, meta), { type });
    }
    visitors.addsGenericOverridden = addsGenericOverridden;
    function addsTypeClassification(meta, ctx) {
        const get = (type) => {
            if (type === undefined)
                return undefined;
            else if (Array.isArray(type))
                return "Array";
            else if (helpers_1.metadata.isCustomClass(type))
                return "Class";
            else
                return "Primitive";
        };
        if (meta.kind === "Method")
            return Object.assign(Object.assign({}, meta), { typeClassification: get(meta.returnType) });
        else if (meta.kind === "Property" || meta.kind === "Parameter")
            return Object.assign(Object.assign({}, meta), { typeClassification: get(meta.type) });
        else if (meta.kind === "Class")
            return Object.assign(Object.assign({}, meta), { typeClassification: "Class" });
        return meta;
    }
    visitors.addsTypeClassification = addsTypeClassification;
    function addsParameterProperties(meta, ctx) {
        if (helpers_1.metadata.isParameterProperties(meta) && ctx.parent.kind === "Class") {
            const isParamProp = ctx.parent.decorators.some((x) => x.type === "ParameterProperties");
            return !!isParamProp ? meta : undefined;
        }
        return meta;
    }
    visitors.addsParameterProperties = addsParameterProperties;
    function removeIgnored(meta, ctx) {
        if (meta.kind === "Property" || meta.kind === "Method") {
            const decorator = meta.decorators.find((x) => x.kind === "Ignore");
            return !decorator ? meta : undefined;
        }
        return meta;
    }
    visitors.removeIgnored = removeIgnored;
})(visitors || (visitors = {}));
exports.visitors = visitors;
// --------------------------------------------------------------------- //
// ------------------------------ WALKERS ------------------------------ //
// --------------------------------------------------------------------- //
/**
 * Walk into type member metadata (properties, parameters, methods, ctor etc)
 * @param meta type metadata
 * @param ctx traversal context
 */
function walkMetadataMembers(meta, ctx) {
    // apply visitor for each metadata traversed
    const result = ctx.visitor(meta, ctx);
    for (const key in result) {
        // walk into type metadata members specified
        if (["parameters", "properties", "methods", "ctor"].some(x => x === key)) {
            const item = result[key];
            if (Array.isArray(item)) {
                const node = item.map((x, i) => walkMetadataMembers(x, Object.assign(Object.assign({}, ctx), { parent: result })));
                result[key] = node.filter(x => !!x);
            }
            else {
                const node = walkMetadataMembers(item, Object.assign(Object.assign({}, ctx), { parent: item }));
                result[key] = node;
            }
        }
    }
    return result;
}
function walkMembers(type, visitor, classPath) {
    const rawMeta = parser_1.parseClass(type);
    return walkMetadataMembers(rawMeta, { visitor, parent: rawMeta, target: type, classPath });
}
/**
 * Walk into type super class
 * @param type type to reflect
 */
function walkClass(type, ctx) {
    // walk first into the parent members
    const parent = Object.getPrototypeOf(type);
    if (parent.prototype) {
        // walk the super class member first
        const parentMeta = walkClass(parent, Object.assign(Object.assign({}, ctx), { classPath: ctx.classPath.concat(type) }));
        // then walk the current type members
        const childMeta = walkMembers(type, ctx.visitor, ctx.classPath);
        // merge current type and super class members
        return extends_1.extendsMetadata(childMeta, parentMeta);
    }
    else {
        return walkMembers(type, ctx.visitor, ctx.classPath);
    }
}
exports.walkClass = walkClass;
